/*
 *  ALTerator - ALT Linux configuration project
 *
 *  Copyright (c) 2004,2005 ALT Linux Ltd.
 *  Copyright (c) 2004,2005 Stanislav Ievlev
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <list>
#include <string>
#include <stdexcept>

#include <command_io.hh>
#include <bus_io.hh>

struct node
{
    typedef std::list<std::string> child_list;

    node(): root_(true), visited_(false)
    {}

    //fixme: add merging here

    command    cmd_; /**< woobus command */
    bool       root_; /**< flag: are this node without parents */
    bool       visited_; /**< flag: are this node was visited */
    child_list childs_; /**< node childs */
};

typedef std::map<std::string,node> cmd_graph;
typedef std::list<command> cmd_list;

struct tree_walk
{
	tree_walk(cmd_graph& g,std::ostream &os):g_(g),os_(os) {}

	void operator()(const std::string& name)
	{
		node &n = g_[name];
		if (n.visited_) return;
		n.visited_ = true;
		
		if (!n.childs_.empty())
			std::for_each(n.childs_.begin(),
				      n.childs_.end(),
				      tree_walk(g_,os_));
		os_<<n.cmd_;
	}
	private:
		cmd_graph&	g_;
		std::ostream&	os_;
};

struct root_walk
{
	root_walk(cmd_graph& g,std::ostream& os):w_(g,os) {}
	void operator()(const cmd_graph::value_type& v)
	{
		if (v.second.root_) w_(v.first);
	}
	private:
		tree_walk w_;
};

//search string "target" in graph begining search at "start" node
struct graph_search
{
	graph_search(cmd_graph& g,const std::string& t):g_(g),t_(t) {}
	
	bool operator()(const std::string& s)
	{
		node& n = g_[s]; //start node
		if (s == t_) return true;
		else
			return std::find_if(n.childs_.begin(),
		                            n.childs_.end(),
					    graph_search(g_,t_)) != n.childs_.end();
	}
	private:
		cmd_graph& g_;
		const std::string& t_;
};


struct search_cmd
{
	search_cmd(const command& c): c_(c),res_(false) {}

	void operator()(const std::string& name,const std::string& value)
	{
		if (res_) return;
		res_ =  ((c_.exists(name) && c_(name) != value));
	}
	operator bool() const { return res_; }

	const command& c_;
	bool res_;
};


bool search_opts(command& cmd1,command& cmd2)
{
	return cmd1.for_each(search_cmd(cmd2));
}

bool process(std::istream& is,std::ostream& os)
{
	cmd_graph graph;
	
	command curr,prereq,error("/error/");
	while (is>>curr)
	{
		if (!curr.name().empty()) //skip empty commands
		{
			if (graph.find(curr.name()) == graph.end())
			{//create new node
				node n;
				n.cmd_ = curr;
	  			graph[curr.name()] = n;
			}
			else
			{//merge existing node with new
				if (search_opts(graph[curr.name()].cmd_,curr))
				{//check for the same options
					os<<error("reason","option overloading");
					return false;
				}
				graph[curr.name()].cmd_.add(curr);
			}

			if (!prereq.name().empty() && (prereq.name() != curr.name()))
			{
				if (graph_search(graph,curr.name())(prereq.name()))
				{
					os<<error("reason","loop was detected on pair "+
							   curr.name()+" with prereq "+prereq.name());
					return false;
				}
				graph[curr.name()].childs_.push_back(prereq.name());
				graph[prereq.name()].root_ = false;
			}
		}
		prereq = curr;
	}
	std::for_each(graph.begin(),graph.end(),root_walk(graph,os));
	return true;
}

void bus_main(std::istream& is,std::ostream& os,std::ostream& log)
{
    while (is)
    {
	std::string direction;
	std::getline(is,direction);
	if (">>" == direction)
	{//read input commands
	    std::istringstream question(receive(is,os,log));
	    std::ostringstream answer;
	    bool res = process(question,answer);
	    send(res?">>":"<<",answer.str(),is,os,log);//send result to other clients
	}
	else if ("<<" == direction)
	{
	    os<<"<>"<<std::endl;
	}
	else
	    log << "--invalid address line: \""
	    <<direction << "\"" << std::endl;
    }
}



int main()
{
    bus_main(std::cin,std::cout,std::cerr);
}
